from investos.portfolio.constraint_model.base_constraint import BaseConstraint


class LongOnlyConstraint(BaseConstraint):
    """
    A constraint that enforces no short positions. Including no short cash position.

    Parameters
    ----------
    **kwargs :
        Additional keyword arguments.
    """

    def __init__(self, **kwargs):
        pass

    def weight_expr(self, t, w_plus, z, v):
        """
        Returns a series of holding constraints.

        Parameters
        ----------
        t : datetime
            The current time.

        w_plus : series
            Portfolio weights after trades z.

        z : series
            Trades for period t

        v : float
            Value of portfolio at period t

        Returns
        -------
        series
            The holding constraints based on the no short positions (including no short cash position) constraint.
        """
        return w_plus >= 0.0


class LongCashConstraint(BaseConstraint):
    """
    A constraint that enforces no short cash positions.

    Parameters
    ----------
    **kwargs :
        Additional keyword arguments.
    """

    def __init__(self, **kwargs):
        pass

    def weight_expr(self, t, w_plus, z, v):
        """
        Returns a series of holding constraints.

        Parameters
        ----------
        t : datetime
            The current time.

        w_plus : series
            Portfolio weights after trades z.

        z : series
            Trades for period t

        v : float
            Value of portfolio at period t

        Returns
        -------
        series
            The holding constraints based on the no short cash positions constraint.
        """
        return w_plus[-1] >= 0.0


class EqualLongShortConstraint(BaseConstraint):
    """
    A constraint that enforces equal long and short exposure.

    Parameters
    ----------
    **kwargs :
        Additional keyword arguments.
    """

    def __init__(self, **kwargs):
        pass

    def weight_expr(self, t, w_plus, z, v):
        """
        Returns a series of holding constraints.

        Parameters
        ----------
        t : datetime
            The current time.

        w_plus : series
            Portfolio weights after trades z.

        z : series
            Trades for period t

        v : float
            Value of portfolio at period t

        Returns
        -------
        series
            The holding constraints based on the equal long and short exposure constraint.
        """
        return sum(w_plus[:-1]) == 0.0



import collections
import datetime as dt

import numpy as np
import pandas as pd

from investos.portfolio.result.save_result import SaveResult
from investos.util import clip_for_dates


class BaseResult(SaveResult):
    """The `Result` class captures portfolio data and performance for each asset and period over time.

    Instances of this object are called by the :py:meth:`investos.portfolio.controller.Controller.generate_positions` method.
    """

    def __init__(self, start_date, end_date):
        self.start_date = start_date
        self.end_date = end_date

    def save_data(self, name: str, t: dt.datetime, entry: pd.Series) -> None:
        """Save `entry` on `Result` object, a (pandas) Series of data as `name` for datetime `t`.

        Parameters
        ----------
        name : str
            The name `entry` is saved under in this `Result` object.
        t : datetime.datetime
            The datetime `entry` is saved under in this `Result` object.
        entry : pandas.Series
            A series of values - for a collection of assets / stocks / tickers at a specific point in time.
        """
        try:
            getattr(self, name).loc[t] = entry
        except AttributeError:
            setattr(
                self,
                name,
                (pd.Series if np.isscalar(entry) else pd.DataFrame)(
                    index=[t], data=[entry]
                ),
            )

    def save_position(self, t: dt.datetime, u: pd.Series, h_next: pd.Series) -> None:
        """
        Save data `u` and `h_next` related to position for datetime `t` on `Result` object.

        Parameters
        ----------
        t : datetime.datetime
            The datetime for associated trades `u` and t + 1 holdings `h_next`.
        u : pandas.Series
            Trades (as values) for period `t`.
        h_next : pandas.Series
            Holdings at beginning of period t + 1, after trades `u` and returns for period `t`.
        """
        self.save_data("u", t, u)
        self.save_data("h_next", t, h_next)

    @property
    def summary(self) -> None:
        """Outputs a string summary of backtest result properties and performance
        (e.g. :py:attr:`~investos.portfolio.result.base_result.BaseResult.num_periods`, :py:attr:`~investos.portfolio.result.base_result.BaseResult.sharpe_ratio`, :py:attr:`~investos.portfolio.result.base_result.BaseResult.max_drawdown`, etc.).
        """
        print(self._summary_string())

    def _summary_string(self) -> str:
        """Returns a string summary of backtest result properties and performance
        (e.g. :py:attr:`~investos.portfolio.result.base_result.BaseResult.num_periods`, :py:attr:`~investos.portfolio.result.base_result.BaseResult.sharpe_ratio`, :py:attr:`~investos.portfolio.result.base_result.BaseResult.max_drawdown`, etc.).

        Do not call directly; call :py:attr:`~investos.portfolio.result.base_result.BaseResult.summary` instead.
        """
        data = collections.OrderedDict(
            {
                "Initial timestamp": self.h.index[0],
                "Final timestamp": self.h.index[-1],
                "Total portfolio return (%)": str(round(self.total_return * 100, 2))
                + "%",
                "Annualized portfolio return (%)": str(
                    round(self.annualized_return * 100, 2)
                )
                + "%",
                "Annualized excess portfolio return (%)": str(
                    round(self.annualized_excess_return * 100, 2)
                )
                + "%",
                "Annualized excess risk (%)": str(
                    round(self.excess_risk_annualized * 100, 2)
                )
                + "%",
                "Information ratio (x)": str(round(self.information_ratio, 2)) + "x",
                "Annualized risk over risk-free (%)": str(
                    round(self.risk_over_cash_annualized * 100, 2)
                )
                + "%",
                "Sharpe ratio (x)": str(round(self.sharpe_ratio, 2)) + "x",
                "Max drawdown (%)": f"{round(self.max_drawdown * 100, 2)}%",
                "Annual turnover (x)": str(round(self.annual_turnover, 2)) + "x",
                "Portfolio hit rate (%)": f"{round(self.portfolio_hit_rate * 100, 2)}%",
            }
        )

        return pd.Series(data=data).to_string(float_format="{:,.3f}".format)

    @property
    @clip_for_dates
    def h(self) -> pd.DataFrame:
        """Returns a pandas Dataframe of asset holdings (`h`) at the beginning of each datetime period."""
        tmp = self.h_next.copy()
        tmp = self.h_next.shift(1)  # Shift h_next to h timing
        return tmp[1:]

    @property
    @clip_for_dates
    def trades(self) -> pd.DataFrame:
        "Returns a pandas Series of trades (u)."
        return self.u

    @property
    def num_periods(self) -> int:
        """Number of periods in backtest. Note that the starting position (at t=0) does not count as a period."""
        return self.h.shape[0]

    @property
    def v(self) -> pd.Series:
        """Returns a pandas Series for the value (`v`) of the portfolio for each datetime period."""
        return self.h.sum(axis=1)

    @property
    def v_with_benchmark(self) -> pd.Series:
        """Returns a pandas Dataframe with simulated portfolio and benchmark values."""
        return pd.DataFrame({"portfolio": self.v, "benchmark": self.benchmark_v})

    @property
    def returns(self) -> pd.Series:
        """Returns a pandas Series of the returns for each datetime period (vs the previous period)."""
        val = self.v
        return pd.Series(
            data=val.values[1:] / val.values[:-1] - 1, index=val.index[1:]
        ).dropna()

    @property
    def portfolio_hit_rate(self):
        return (self.returns > 0).mean()

    @property
    def total_return(self) -> float:
        """Returns a float representing the total return for the entire period under review."""
        return self.v[-1] / self.v[0] - 1

    @property
    def total_benchmark_return(self) -> float:
        """Returns a float representing the return over benchmark for the entire period under review."""
        return self.benchmark_v[-1] / self.benchmark_v[0] - 1

    @property
    def total_risk_free_return(self) -> float:
        return (1 + self.risk_free_returns).cumprod().iloc[-1] - 1

    @property
    def total_excess_return(self) -> float:
        """Returns a float representing the total return for the entire period under review."""
        return self.total_return - self.total_benchmark_return

    @property
    def total_return_over_cash(self) -> float:
        """Returns a float representing the total returns over cash for the entire period under review."""
        return self.total_return - self.total_risk_free_return

    @property
    def annualized_return(self) -> float:
        """Returns a float representing the annualized return of the entire period under review. Uses beginning and ending portfolio values for the calculation (value @ t[-1] and value @ t[0]), as well as the number of years in the forecast."""
        return ((self.total_return + 1) ** (1 / self.years_forecast)) - 1

    @property
    def annualized_benchmark_return(self) -> float:
        """Returns a float representing the annualized benchmark return of the entire period under review. Uses beginning and ending portfolio values for the calculation (value @ t[-1] and value @ t[0]), as well as the number of years in the forecast."""
        return ((self.total_benchmark_return + 1) ** (1 / self.years_forecast)) - 1

    @property
    def excess_returns(self) -> pd.Series:
        """Returns a pandas Series of returns in excess of the benchmark."""
        return (self.returns - self.benchmark_returns).dropna()

    @property
    def returns_over_cash(self) -> pd.Series:
        """Returns a pandas Series of returns in excess of risk free returns."""
        return (self.returns - self.risk_free_returns).dropna()

    @property
    def annualized_excess_return(self) -> float:
        """Returns a float representing the annualized excess return of the entire period under review. Uses beginning and ending portfolio values for the calculation (value @ t[-1] and value @ t[0]), as well as the number of years in the forecast."""
        return ((self.total_excess_return + 1) ** (1 / self.years_forecast)) - 1

    @property
    def annualized_return_over_cash(self) -> float:
        """Returns a float representing the annualized return over cash of the entire period under review. Uses beginning and ending portfolio values for the calculation (value @ t[-1] and value @ t[0]), as well as the number of years in the forecast."""
        return ((self.total_return_over_cash + 1) ** (1 / self.years_forecast)) - 1

    @property
    def excess_risk_annualized(self) -> pd.Series:
        """Returns a pandas Series of risk in excess of the benchmark."""
        return self.excess_returns.std() * np.sqrt(self.ppy)

    @property
    def risk_over_cash_annualized(self) -> pd.Series:
        """Returns a pandas Series of risk in excess of the risk free rate."""
        return self.returns_over_cash.std() * np.sqrt(self.ppy)

    @property
    @clip_for_dates
    def benchmark_returns(self) -> pd.Series:
        if not hasattr(self, "benchmark"):
            self.benchmark = self.optimizer.actual["return"]["cash"]

        return self.benchmark

    @property
    @clip_for_dates
    def risk_free_returns(self) -> pd.Series:
        if not hasattr(self, "risk_free"):
            self.risk_free = self.optimizer.actual["return"]["cash"]

        return self.risk_free

    @property
    def benchmark_v(self) -> pd.Series:
        """Returns series of simulated portfolio values, if portfolio was invested 100% in benchmark at time 0"""
        benchmark_factors = self.benchmark_returns + 1
        benchmark_factors[0] = 1  # No returns for period 0

        return (
            benchmark_factors.cumprod() * self.v[0]
        )  # Calculate values if initial portfolio value was invested 100% in benchmark

    @property
    def years_forecast(self) -> float:
        """Returns a float representing the number of years in the backtest period.
        Calculated as (datetime @ t[-1] - datetime @ t[0]) / datetime.timedelta(days=365.25)
        """
        return (self.v.index[-1] - self.v.index[0]) / dt.timedelta(days=365.25)

    @property
    def ppy(self) -> float:
        """Returns a float representing the number of periods per year in the backtest period.
        Calculated as :py:attr:`~investos.portfolio.result.base_result.BaseResult.num_periods` / :py:attr:`~investos.portfolio.result.base_result.BaseResult.years_forecast`
        """
        return self.num_periods / self.years_forecast

    @property
    def information_ratio(self, use_annualized_inputs=True) -> float:
        """Returns a float representing the (annualized) Information Ratio of the portfolio.

        Ratio is calculated as mean of :py:attr:`~investos.portfolio.result.base_result.base_result.BaseResult.excess_returns` / standard deviation of :py:attr:`~investos.portfolio.result.base_result.BaseResult.excess_returns`. Annualized by multiplying ratio by square root of periods per year (:py:attr:`~investos.portfolio.result.base_result.BaseResult.ppy`).
        """
        if use_annualized_inputs:
            return self.annualized_excess_return / self.excess_risk_annualized
        else:
            return (
                np.sqrt(self.ppy)
                * np.mean(self.excess_returns)
                / np.std(self.excess_returns)
            )

    @property
    def sharpe_ratio(self, use_annualized_inputs=True) -> float:
        """Returns a float representing the (annualized) Sharpe Ratio of the portfolio.

        Ratio is calculated as mean of :py:attr:`~investos.portfolio.result.base_result.base_result.BaseResult.excess_returns` / standard deviation of :py:attr:`~investos.portfolio.result.base_result.BaseResult.excess_returns`. Annualized by multiplying ratio by square root of periods per year (:py:attr:`~investos.portfolio.result.base_result.BaseResult.ppy`).
        """
        if use_annualized_inputs:
            return self.annualized_return_over_cash / self.risk_over_cash_annualized
        else:
            return (
                np.sqrt(self.ppy)
                * np.mean(self.returns_over_cash)
                / np.std(self.returns_over_cash)
            )

    @property
    def turnover(self):
        """Turnover ||u_t||_1/v_t"""
        noncash_trades = self.trades.drop(["cash"], axis=1)
        return np.abs(noncash_trades).sum(axis=1) / self.v

    @property
    def leverage(self):
        """Turnover ||u_t||_1/v_t"""
        noncash_h = self.h.drop(["cash"], axis=1)
        return np.abs(noncash_h).sum(axis=1) / self.v

    @property
    def long_leverage(self):
        """Turnover ||u_t||_1/v_t"""
        noncash_h = self.h.drop(["cash"], axis=1)
        return np.abs(noncash_h[noncash_h > 0]).sum(axis=1) / self.v

    @property
    def short_leverage(self):
        """Turnover ||u_t||_1/v_t"""
        noncash_h = self.h.drop(["cash"], axis=1)
        return np.abs(noncash_h[noncash_h < 0]).sum(axis=1) / self.v

    @property
    def annual_turnover(self):
        return self.turnover.mean() * self.ppy

    @property
    def max_drawdown(self):
        """The maximum peak to trough drawdown in percent."""
        val_arr = self.v.values
        max_dd_so_far = 0
        cur_max = val_arr[0]
        for val in val_arr[1:]:
            if val >= cur_max:
                cur_max = val
            elif (cur_max - val) / cur_max > max_dd_so_far:
                max_dd_so_far = (cur_max - val) / cur_max
        return max_dd_so_far

    def hit_rate(self, returns_df):
        h = self.h
        if "cash" in list(h.columns):
            h = h.drop(["cash"], axis=1)
        if "cash" in list(returns_df.columns):
            returns_df = returns_df.drop(["cash"], axis=1)

        h = h.iloc[1:].round(decimals=0).replace(-0.0, 0.0)
        returns_df = returns_df[returns_df.index.isin(h.index)]
        sign_agree_df = (np.sign(h) * np.sign(returns_df)).fillna(0).astype(int)

        hit = sign_agree_df[sign_agree_df == 1].sum(axis=1)
        hit_attempt = sign_agree_df[sign_agree_df.abs() == 1].abs().sum(axis=1)
        return hit / hit_attempt



import cvxpy as cvx
import numpy as np
import pandas as pd

from investos.portfolio.risk_model import BaseRisk


class StatFactorRisk(BaseRisk):
    """PCA-factor based risk model.

    The only requirement of custom risk models is that they implement a `_estimated_cost_for_optimization` method.

    Note: risk models are like cost models, except they return 0 for their `value_expr` method (because they only influence optimization weights, not actual cash costs).
    """

    def __init__(self, n_factors=5):
        self.n = n_factors

        self.optimizer = None  # Set during Controller initialization

        self.factor_variance = None
        self.factor_loadings = None
        self.idiosyncratic_variance = None

        super().__init__()

    def _estimated_cost_for_optimization(self, t, w_plus, z, value):
        """Optimization (non-cash) cost penalty for assuming associated asset risk.

        Used by optimization strategy to determine trades.

        Not used to calculate simulated costs for backtest performance.
        """
        if (
            self.factor_variance is None
            or self.factor_loadings is None
            or self.idiosyncratic_variance is None
        ):
            self.create_risk_model()

        self.expression = cvx.sum(
            cvx.multiply(self.idiosyncratic_variance, cvx.square(w_plus))
        ) + cvx.sum(cvx.multiply(w_plus.T @ self.factor_loadings, self.factor_variance))

        return self.expression, []

    def create_risk_model(self):
        df = self.optimizer.actual["return"]
        forecast_start_d = self.optimizer.forecast["date"]["start"]
        df = df[
            (df.index < forecast_start_d)
            & (
                df.index
                >= pd.to_datetime(forecast_start_d)
                - pd.Timedelta("730 days")  # 2 years
            )
        ]

        covariance_matrix = df.cov().dropna().values
        eigenvalue, eigenvector = np.linalg.eigh(covariance_matrix)

        self.factor_variance = eigenvalue[-self.n :]
        self.factor_loadings = pd.DataFrame(
            data=eigenvector[:, -self.n :], index=df.columns
        )
        self.factor_loadings.iloc[-1] = 0  # For cash
        self.idiosyncratic_variance = pd.Series(
            data=np.diag(
                eigenvector[:, : -self.n]
                @ np.diag(eigenvalue[: -self.n])
                @ eigenvector[:, : -self.n].T
            ),
            index=df.columns,
        )
        self.idiosyncratic_variance = self.idiosyncratic_variance.copy()
        self.idiosyncratic_variance.iloc[-1] = 0



import cvxpy as cvx
import numpy as np

from investos.portfolio.constraint_model.base_constraint import BaseConstraint
from investos.util import values_in_time


class MaxTradeConstraint(BaseConstraint):
    """
    A constraint that enforces a limit on the (absolute) max trade size (as a weight, or fraction, of daily volume).

    Parameters
    ----------
    limit : float, optional
        The limit on the (absolute) max trade size (as a weight, or fraction, of volume for period `t`).

    **kwargs :
        Additional keyword arguments.
    """

    def __init__(self, limit: float = 1.0, **kwargs):
        self.limit = limit

    def weight_expr(self, t, w_plus, z, v):
        """
        Returns a series of holding constraints.

        Parameters
        ----------
        t : datetime
            The current time.

        w_plus : series
            Portfolio weights after trades z.

        z : series
            Trades for period t

        v : float
            Value of portfolio at period t

        Returns
        -------
        series
            The holding constraints based on the max trade constraint.
        """
        return (
            cvx.abs(z[:-1]) * v
            <= np.array(
                values_in_time(self.optimizer.forecast["volume"], t)
                * values_in_time(self.optimizer.forecast["price"], t)
            )
            * self.limit
        )



import pandas as pd

from investos.portfolio.result.base_result import BaseResult


class WeightsResult(BaseResult):
    """For generating backtest results from portfolio weights and historical returns.

    In this model, trades for each period happen after returns.
    """

    def __init__(
        self, initial_weights, trade_weights, returns, aum=100_000_000, *args, **kwargs
    ):
        self.set_h_next(initial_weights, trade_weights, returns, aum)
        self.start_date = kwargs.get("start_date", trade_weights.index[0])
        self.end_date = kwargs.get("end_date", trade_weights.index[-1])
        self.risk_free = kwargs.get("risk_free", pd.Series(0.0, index=returns.index))
        self.benchmark = kwargs.get("benchmark", pd.Series(0.0, index=returns.index))

    def set_h_next(self, initial_weights, trade_weights, returns, aum):
        print(
            "Calculating holding values from trades, returns, and initial weights and AUM..."
        )

        h = initial_weights
        for t in trade_weights.index:
            r = returns.loc[t]
            u = trade_weights.loc[t]
            h_next = (h + u) * (1.0 + r)
            self.save_position(t, u, h_next)
            h = h_next

        self.u *= aum
        self.h_next *= aum

        print("Done calculations.")



import copy
import datetime as dt

import pandas as pd


class BaseCost:
    """Base cost model for InvestOS.
    Other cost models should subclass BaseCost.
    The only requirement of custom cost models is that they (re)implement :py:meth:`~investos.portfolio.cost_model.base_cost.BaseCost.value_expr`.
    """

    def __init__(self):
        self.optimizer = None  # Set during Optimizer initialization
        self.gamma = 1  # Can change without setting directly as: gamma * BaseCost(). Note that gamma doesn't impact actual costs in backtester / simulated performance, just trades in optimization strategy.

    def weight_expr(self, t, w_plus, z, value):
        cost, constraints = self._estimated_cost_for_optimization(t, w_plus, z, value)
        return self.gamma * cost, constraints

    def actual_cost(self, t: dt.datetime, h_plus: pd.Series, u: pd.Series) -> pd.Series:
        """Method that calculates per-period costs given period `t` holdings and trades.

        Parameters
        ----------
        t : datetime.datetime
            The datetime for associated trades `u` and holdings plus trades `h_plus`.
        h_plus : pandas.Series
            Holdings at beginning of period t, plus trades for period `t` (`u`). Same as `u` + `h` for `t`.
        u : pandas.Series
            Trades (as values) for period `t`.
        """
        raise NotImplementedError

    def __mul__(self, other):
        """Read the gamma parameter as a multiplication; so you can change self.gamma without setting it directly as: gamma * BaseCost()"""
        newobj = copy.copy(self)
        newobj.gamma *= other
        return newobj

    def __rmul__(self, other):
        """Read the gamma parameter as a multiplication; so you can change self.gamma without setting it directly as: gamma * BaseCost()"""
        return self.__mul__(other)

    def metadata_dict(self):
        metadata_dict = {}

        metadata_dict["gamma"] = self.gamma

        if getattr(self, "price_movement_sensitivity", None):
            metadata_dict["price_movement_sensitivity"] = self.limit

        return metadata_dict



import datetime as dt

import pandas as pd

from investos.portfolio.cost_model import BaseCost


class BaseRisk(BaseCost):
    """Base risk model for InvestOS.

    The only requirement of custom risk models is that they implement a `_estimated_cost_for_optimization` method.

    Note: risk models are like cost models, except they return 0 for their `value_expr` method (because they only influence optimization weights, not actual cash costs).
    """

    def __init__(self):
        self.optimizer = None  # Set during Controller initialization

        super().__init__()

    def actual_cost(self, t: dt.datetime, h_plus: pd.Series, u: pd.Series) -> pd.Series:
        """Method that calculates per-period costs given period `t` holdings and trades.

        ALWAYS 0 FOR RISK MODELS; DO NOT OVERRIDE. RISK DOESN'T HAVE A CASH COST, IT ONLY AFFECTS OPTIMIZED ASSET WEIGHTS GIVEN `_estimated_cost_for_optimization` RISK (UTILITY) PENALTY.
        """
        return 0

    def _estimated_cost_for_optimization(self, t, w_plus, z, value):
        """Optimization (non-cash) cost penalty for assuming associated asset risk.

        Used by optimization strategy to determine trades.

        Not used to calculate simulated costs for backtest performance.
        """
        raise NotImplementedError



import requests

import investos


class SaveResult:
    def save(
        self,
        description,
        tags=[],
        team_ids=[],
        strategy=None,
        forecast_ids=[],
    ):
        self.api_key = investos.api_key
        self.api_endpoint = investos.api_endpoint
        self.request_headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        self.tags = tags
        self.team_ids = team_ids

        self.save_backtest(description, strategy, forecast_ids)
        self.save_backtest_charts()

    def save_backtest(self, description, strategy, forecast_ids):
        json_body = {
            "backtest": {
                "description": description,
                "performance_summary": {
                    "total_return": self.total_return,
                    "annualized_return": self.annualized_return,
                    "sharpe_ratio": self.sharpe_ratio,
                    "max_drawdown": self.max_drawdown,
                    "annual_turnover": self.annual_turnover,
                    "portfolio_hit_rate": self.portfolio_hit_rate,
                },
                "tags": self.tags,
                "team_ids": self.team_ids,
                "portfolio_construction": strategy and strategy.metadata_dict(),
                "forecast_ids": forecast_ids,
            }
        }

        response = requests.post(
            f"{self.api_endpoint}/backtests",
            headers=self.request_headers,
            json=json_body,
        )

        if (
            response.status_code // 100 == 2
        ):  # Check if the status code is in the 200 range
            self.backtest_id = response.json().get("id")
            print(f"Backtest {self.backtest_id} saved. Creating charts.")
        else:
            print(
                f"Backtest API request failed with status code: {response.status_code}"
            )

    def save_backtest_charts(self):
        self.save_chart_historical_value()
        self.save_chart_historical_returns()
        self.save_chart_historical_leverage()

    def save_chart_historical_value(self):
        json_body = {
            "chart": {
                "title": "Value evolution",
                "chartable_type": "Backtest",
                "chartable_id": self.backtest_id,
                "chart_traces": [
                    {
                        "x_name": "Dates",
                        "y_name": "Portfolio",
                        "x_values": [str(el) for el in self.v.index],
                        "y_values": list(self.v.values),
                    },
                    {
                        "x_name": "Dates",
                        "y_name": "Benchmark",
                        "x_values": [str(el) for el in self.benchmark_v.index],
                        "y_values": list(self.benchmark_v.values),
                    },
                ],
            }
        }

        self._save_chart(json_body)

    def save_chart_historical_returns(self):
        num_periods = self.v.shape[0] - 1
        rolling_num_a = min(20, num_periods)
        rolling_num_b = min(60, num_periods)
        rolling_num_c = min(252, num_periods)

        json_body = {
            "chart": {
                "title": "Rolling return evolution",
                "chartable_type": "Backtest",
                "chartable_id": self.backtest_id,
                "chart_traces": [
                    {
                        "x_name": "Dates",
                        "y_name": f"{rolling_num_a} trading days",
                        "x_values": [str(el) for el in self.v.index],
                        "y_values": list(
                            self.v.pct_change(periods=rolling_num_a)
                            .fillna(method="bfill")
                            .values
                        ),
                    },
                    {
                        "x_name": "Dates",
                        "y_name": f"{rolling_num_b} trading days",
                        "x_values": [str(el) for el in self.v.index],
                        "y_values": list(
                            self.v.pct_change(periods=rolling_num_b)
                            .fillna(method="bfill")
                            .values
                        ),
                    },
                    {
                        "x_name": "Dates",
                        "y_name": f"{rolling_num_c} trading days",
                        "x_values": [str(el) for el in self.v.index],
                        "y_values": list(
                            self.v.pct_change(periods=rolling_num_c)
                            .fillna(method="bfill")
                            .values
                        ),
                    },
                ],
            }
        }

        self._save_chart(json_body)

    def save_chart_historical_leverage(self):
        json_body = {
            "chart": {
                "title": "Leverage evolution",
                "chartable_type": "Backtest",
                "chartable_id": self.backtest_id,
                "chart_traces": [
                    {
                        "x_name": "Dates",
                        "y_name": "Long",
                        "x_values": [str(el) for el in self.long_leverage.index],
                        "y_values": list(self.long_leverage.values),
                    },
                    {
                        "x_name": "Dates",
                        "y_name": "Short",
                        "x_values": [str(el) for el in self.short_leverage.index],
                        "y_values": list(self.short_leverage.values),
                    },
                ],
            }
        }

        self._save_chart(json_body)

    def save_chart_hit_rate(self, returns_df):
        hit_rate = self.hit_rate(returns_df)
        hit_rate_mean = hit_rate.mean()
        hit_rate_rolling = hit_rate.rolling(window=60, min_periods=1).mean()

        json_body = {
            "chart": {
                "title": "Holdings hit rate",
                "chartable_type": "Backtest",
                "chartable_id": self.backtest_id,
                "chart_traces": [
                    {
                        "x_name": "Dates",
                        "y_name": "Current period",
                        "x_values": [str(el) for el in hit_rate.index],
                        "y_values": list(hit_rate.values),
                    },
                    {
                        "x_name": "Dates",
                        "y_name": "Rolling 60 periods",
                        "x_values": [str(el) for el in hit_rate_rolling.index],
                        "y_values": list(hit_rate_rolling.values),
                    },
                    {
                        "x_name": "Dates",
                        "y_name": "Mean",
                        "x_values": [str(el) for el in hit_rate.index],
                        "y_values": [hit_rate_mean for el in hit_rate.index],
                    },
                ],
            }
        }

        self._save_chart(json_body)

    def _save_chart(self, json_body):
        response = requests.post(
            f"{self.api_endpoint}/charts/create_or_update",
            headers=self.request_headers,
            json=json_body,
        )

        if (
            response.status_code // 100 == 2
        ):  # Check if the status code is in the 200 range
            chart_id = response.json().get("id")
            print(f"Chart {chart_id} saved.")
        else:
            print(f"Chart creation failed with status code: {response.status_code}")



import datetime as dt
import statistics
from typing import Callable

import numpy as np
import pandas as pd

import investos.portfolio.result as result
import investos.util as util
from investos.portfolio.strategy import BaseStrategy, RankLongShort


class Controller:
    """Container class that runs backtests using passed-in portfolio optimization `strategy` (see :py:class:`~investos.portfolio.strategy.base_strategy.BaseStrategy`), then saves results into passed-in `result` (see :py:class:`~investos.backtest.result.Result`) class.

    Parameters
    ----------
    df_forecast : pd.DataFrame
        DataFrame of forecast asset returns for backtest period.

        Expected columns:
        asset (unique ID/ticker),
        date (datetime),
        return (float)

    df_actual : pd.DataFrame
        DataFrame of actual asset returns for backtest period. Used to create forecasts for std_dev, spread, and volume if forecasts not given.

        Expected columns:
        asset (unique ID/ticker),
        date (datetime),
        price (float),
        return (float),
        volume (of shares traded, float),
        spread (float)

    strategy : :py:class:`~investos.portfolio.strategy.base_strategy.BaseStrategy`
        Optimization strategy used by backtester. Used to determine what trades to make at each forecast time period.

    backtest_model : :py:class:`~investos.portfolio.result.BaseResult`
        Stores result from simulated backtest, and contains convenience properties and methods for reporting.

    initial_portfolio : pd.DataFrame, optional
        Initial portfolio values in dollars (or other currency), not in weights.

    aum : float, optional
        Total assets under management, in dollars (or other currency).

    df_categories : pd.DataFrame, optional
        Additional category data for assets. Used by certain optimization strategies.

    config : dict, optional
        Configuration parameters.

    **kwargs :
        Additional keyword arguments.

    Attributes
    ----------
    config : dict
        Configuration parameters after merging with base configuration.

    strategy : :py:class:`~investos.portfolio.strategy.base_strategy.BaseStrategy`
        Optimization strategy used by backtester.

    backtest_model : :py:class:`~investos.portfolio.result.BaseResult`
        Stores result from simulated backtest.

    forecast : dict
        Holds the pivoted and filled forecast DataFrames.

    actual : dict
        Holds the pivoted and filled actual DataFrames, if provided.

    initial_portfolio : pd.DataFrame
        Initial portfolio values, including cash.

    after_step : [Callable]
        Callable objects, that are passed a reference to Controller (i.e. self), that run at end of each step through time

    Methods
    -------
    generate_positions(self):
        Optimizes the portfolio and backtests it. Returns :py:class:`~investos.portfolio.result.BaseResult` object.

    pivot_and_fill(self, df: pd.DataFrame, values: str, columns='asset', index='date', fill_method='bfill'):
        Pivots and fills the DataFrame based on the provided values, columns, and index.

    create_forecast(self, df_forecast: pd.DataFrame, col_name: str = 'std_dev'):
        Creates a forecast DataFrame based on the provided df_forecast and col_name.

    create_price(self, df_return: pd.DataFrame):
        Creates price data based on provided return data and last historical price.

    create_initial_portfolio(self, initial_portfolio, aum):
        Creates the initial portfolio based on provided initial portfolio (or aum value if iniitial portfolio not provided).

    get_actual_positions_for_t(self, h: pd.Series, u: pd.Series, t: dt.Series):
        Gets actual portfolio positions and holdings for time period t, given trades u, and associates cost models and returns.

        Args:
            h: pandas Series object describing current portfolio
            u: n pandas Series vector with stock trades
            t: current datetime

        Returns:
            h_next: pandas Series portfolio after returns propagation (for t to t+1 period)
            u: pandas Series trades vector with simulated cash balance

    get_initial_t(self):
        Gets the initial time period for the backtest.
    """

    BASE_CONFIG = {
        "forecast": {
            "std_dev": {
                "calc_from_n_prev_periods": 100,  # Calc from actual_df if not passed in
            },
            "half_spread": {
                "calc_from_n_prev_periods": 100,  # Calc from actual_df if not passed in
            },
            "volume": {
                "calc_from_n_prev_periods": 100,  # Calc from actual_df if not passed in
            },
        },
        "borrowing": {"interest_rate": 0.005, "short_rate": 0.005},
    }

    def __init__(
        self,
        df_forecast: pd.DataFrame,
        df_actual: pd.DataFrame = None,
        strategy: BaseStrategy = RankLongShort,
        backtest_model: result.BaseResult = None,
        initial_portfolio: pd.DataFrame = None,  # In dollars (or other currency), not in weights
        aum: float = 100_000_000,
        df_categories: pd.DataFrame = None,
        start_date=None,
        end_date=None,
        after_step: [Callable] = [],
        config: dict = {},
        **kwargs,
    ):
        self.config = util.deep_dict_merge(self.BASE_CONFIG, config)

        self._init_strategy(strategy)
        self._init_backtest_model(backtest_model, df_actual)

        self.forecast = {}
        df_forecast = self._clip_forecast_df_for_dates(
            start_date, end_date, df_forecast
        )

        self._init_df_actual(df_actual)
        self._init_df_forecast(df_forecast)

        self._set_cash_return_for_each_period(self.forecast["return"])
        self.create_initial_portfolio(initial_portfolio, aum)
        self._set_references_back_to_optimizer()

        self.after_step = after_step  # Hook for callables at end of each step in t

    def pivot_and_fill(
        self, df, values, columns="asset", index="date", fill_method="bfill"
    ):
        return pd.pivot(df, values=values, columns=columns, index=index).fillna(
            method=fill_method
        )

    def create_forecast(self, df_forecast, col_name="std_dev"):
        if col_name in df_forecast.columns:
            return df_forecast
        elif col_name == "std_dev":
            return df_forecast[["date", "asset"]].merge(
                self.actual["return"][
                    self.actual["return"].index < self.forecast["date"]["start"]
                ]
                .tail(self.config["forecast"][col_name]["calc_from_n_prev_periods"])
                .std()
                .rename(col_name)
                .reset_index(),
                how="left",
                on="asset",
            )
        else:
            return df_forecast[["date", "asset"]].merge(
                self.actual[col_name][
                    self.actual[col_name].index < self.forecast["date"]["start"]
                ]
                .tail(self.config["forecast"][col_name]["calc_from_n_prev_periods"])
                .mean()
                .rename(col_name)
                .reset_index(),
                how="left",
                on="asset",
            )

    def create_price(self, df_return):
        last_historical_prices = self.actual["price"].loc[
            self.actual["price"][
                self.actual["price"].index < self.forecast["date"]["start"]
            ].index.max()
        ]

        # Make geometric - add 1
        df = df_return + 1

        # Geometric cumulative mean
        df = df.cumprod()

        # Multiply geometric return factors by last historical value
        return df * last_historical_prices

    def create_initial_portfolio(self, initial_portfolio, aum):
        if initial_portfolio is None:
            initial_portfolio = pd.Series(index=self.forecast["return"].columns, data=0)
            initial_portfolio["cash"] = aum

        self.initial_portfolio = initial_portfolio

    def generate_positions(self):
        print("Optimizing...")

        self.backtest = self.backtest_model(
            start_date=self.forecast["date"]["start"],
            end_date=self.forecast["date"]["end"],
        )

        # Submit initial position
        t = self.get_initial_t()
        u = pd.Series(index=self.initial_portfolio.index, data=0)  # No trades at time 0
        h_next = self.initial_portfolio  # Includes cash
        self.backtest.save_position(t, u, h_next)

        # Walk through time and calculate future trades, estimated and actual costs and returns, and resulting positions
        for t in self.forecast["return"].index:
            u = self.strategy.generate_trade_list(h_next, t)
            h_next, u = self.get_actual_positions_for_t(h_next, u, t)
            self.backtest.save_position(t, u, h_next)

            for func in self.after_step:  # Run after_step hooks
                func(self, t, u, h_next)

        print("Done simulating.")

        return self.backtest

    def get_actual_positions_for_t(self, h, u, t):
        h_plus = h + u

        costs = [
            cost.actual_cost(t, h_plus=h_plus, u=u) for cost in self.strategy.costs
        ]

        u["cash"] = -sum(u[u.index != "cash"]) - sum(costs)
        h_plus["cash"] = h["cash"] + u["cash"]

        h_next = self.actual["return"].loc[t] * h_plus + h_plus

        zero_threshold = 0.00001
        h_next[np.abs(h_next) < zero_threshold] = 0
        u[np.abs(u) < zero_threshold] = 0

        return h_next, u

    def get_initial_t(self):
        median_time_delta = statistics.median(
            self.forecast["return"].index[1:5] - self.forecast["return"].index[0:4]
        )

        return self.forecast["return"].index[0] - median_time_delta

    def _set_cash_return_for_each_period(self, df):
        df["cash"] = self.config["borrowing"]["interest_rate"]

        df["tmp_date"] = df.index
        df["tmp_date_lagged"] = df["tmp_date"].shift(1)
        df["tmp_date_delta"] = df["tmp_date"] - df["tmp_date_lagged"]
        df["tmp_date_delta_fraction_of_year"] = df["tmp_date_delta"] / dt.timedelta(
            365, 0, 0, 0
        )
        df["cash"] = (1 + df["cash"]) ** df["tmp_date_delta_fraction_of_year"] - 1
        df["cash"] = df["cash"].fillna(method="bfill")

        df.drop(
            columns=[
                "tmp_date",
                "tmp_date_lagged",
                "tmp_date_delta",
                "tmp_date_delta_fraction_of_year",
            ],
            inplace=True,
        )

    def _set_references_back_to_optimizer(self):
        self.strategy.forecast_returns = self.forecast["return"]
        self.strategy.optimizer = self

        for c in self.strategy.costs + (self.strategy.constraints or []):
            c.optimizer = self

    def _init_df_actual(self, df_actual):
        if df_actual is not None:
            self.actual = {}
            self.actual["return"] = self.pivot_and_fill(df_actual, values="return")
            self.actual["price"] = self.pivot_and_fill(df_actual, values="price")
            self.actual["volume"] = self.pivot_and_fill(df_actual, values="volume")
            self.actual["half_spread"] = (
                self.pivot_and_fill(df_actual, values="spread") / 2
            ).rename(columns={"spread": "half_spread"})
            self.actual["std_dev"] = self.pivot_and_fill(
                self.create_forecast(df_actual, "std_dev"), values="std_dev"
            )
            self._set_cash_return_for_each_period(self.actual["return"])

    def _init_backtest_model(self, backtest_model, df_actual):
        if backtest_model is None:
            if df_actual is not None:
                self.backtest_model = result.BaseResult
            else:
                self.backtest_model = result.ForecastResult
        else:
            self.backtest_model = backtest_model  # Not initialized when passed in

        self.backtest_model.optimizer = self

    def _clip_forecast_df_for_dates(self, start_date, end_date, df_forecast):
        self.forecast["date"] = {
            "start": start_date or df_forecast.date.min(),
            "end": end_date or df_forecast.date.max(),
        }
        return df_forecast[
            (df_forecast["date"] >= self.forecast["date"]["start"])
            & (df_forecast["date"] <= self.forecast["date"]["end"])
        ]

    def _init_df_forecast(self, df_forecast):
        self.forecast["return"] = self.pivot_and_fill(df_forecast, values="return")
        self.forecast["std_dev"] = self.pivot_and_fill(
            self.create_forecast(df_forecast, "std_dev"), values="std_dev"
        )
        self.forecast["volume"] = self.pivot_and_fill(
            self.create_forecast(df_forecast, "volume"), values="volume"
        )
        self.forecast["half_spread"] = self.pivot_and_fill(
            self.create_forecast(df_forecast, "half_spread"), values="half_spread"
        )
        self.forecast["price"] = self.create_price(self.forecast["return"])

    def _init_strategy(self, strategy):
        self.strategy = strategy  # Must be initialized first
        if getattr(self.strategy, "risk_model", None):
            self.strategy.costs += [self.strategy.risk_model]



import datetime as dt

import pandas as pd


class BaseStrategy:
    """Base class for an optimization strategy.

    Must implement :py:meth:`~investos.portfolio.strategy.base_strategy.BaseStrategy.generate_trade_list` as per below.

    Attributes
    ----------
    costs : list
        Cost models evaluated during optimization strategy. Defaults to empty list. See :py:class:`~investos.portfolio.cost_model.base_cost.BaseCost` for cost model base class.
    constraints : list
        Constraints applied for optimization strategy. Defaults to empty list. See [TBU] for optimization model base class.
    """

    def __init__(self):
        self.costs = []
        self.constraints = []
        self.risk_model = None

    def _zerotrade(self, holdings):
        return pd.Series(index=holdings.index, data=0.0)

    def generate_trade_list(self, holdings: pd.Series, t: dt.datetime) -> pd.Series:
        """Calculates and returns trade list (in units of currency passed in), given (added) optimization logic.

        Parameters
        ----------
        holdings : pandas.Series
            Holdings at beginning of period `t`.
        t : datetime.datetime
            The datetime for associated holdings `holdings`.
        """
        raise NotImplementedError

    def metadata_dict(self):
        return {
            "strategy": self.__class__.__name__,
            "risk_model": self.risk_model
            and {self.risk_model.__class__.__name__: self.risk_model.metadata_dict()},
            "constraint_models": {
                el.__class__.__name__: el.metadata_dict() for el in self.constraints
            },
            "cost_models": {
                el.__class__.__name__: el.metadata_dict()
                for el in self.costs
                if "Risk" not in el.__class__.__name__
            },
        }



import datetime as dt

import pandas as pd

from investos.portfolio.cost_model import BaseCost, HoldingCost, TradingCost
from investos.portfolio.strategy import BaseStrategy
from investos.util import values_in_time


class RankLongShort(BaseStrategy):
    """Optimization strategy that builds trade list by going long assets with best return forecasts and short stocks with worst return forecasts.


    Attributes
    ----------
    costs : list[:py:class:`~investos.portfolio.cost_model.base_cost.BaseCost`]
        Cost models evaluated during optimization strategy.
    constraints : list[TBU]
        Constraints applied for optimization strategy. Defaults to empty list.
    percent_short : float
        Percent of assets in forecast returns to go short.
    percent_long : float
        Percent of assets in forecast returns to go long.
    leverage : float
        Absolute value of exposure / AUM. Used to calculate holdings.
    n_periods_held : integer
        Number of periods positions held. After n number of periods, positions unwound.
    """

    def __init__(
        self,
        n_periods_held: int = 1,
        leverage: float = 1,
        percent_short: float = 0.25,
        percent_long: float = 0.25,
        costs: [BaseCost] = [],
    ):
        self.forecast_returns = None  # Set by Controller in init
        self.optimizer = None  # Set by Controller in init

        self.costs = costs
        if not self.costs:
            self.costs = [TradingCost(), HoldingCost()]

        self.constraints = []
        self.risk_model = None

        self.percent_short = percent_short
        self.percent_long = percent_long
        self.n_periods_held = n_periods_held
        self.leverage_per_trade = leverage / n_periods_held

    def generate_trade_list(self, holdings: pd.Series, t: dt.datetime) -> pd.Series:
        """Calculates and returns trade list (in units of currency passed in) by going long top :py:attr:`~investos.portfolio.strategy.rank_long_short.RankLongShort.percent_long` assets and short bottom :py:attr:`~investos.portfolio.strategy.rank_long_short.RankLongShort.percent_short` assets.

        Parameters
        ----------
        holdings : pandas.Series
            Holdings at beginning of period `t`.
        t : datetime.datetime
            The datetime for associated holdings `holdings`.
        """
        w = self._get_trade_weights_for_t(holdings, t)
        u = sum(holdings) * w * self.leverage_per_trade

        idx_t = self.forecast_returns.index.get_loc(t)
        if idx_t - self.n_periods_held >= 0:
            # Use holdings_unwind, t_unwind, w_unwind, u_unwind, u_unwind_scaled
            t_unwind = self.forecast_returns.index[idx_t - self.n_periods_held]
            holdings_unwind = self.optimizer.backtest.h.loc[t_unwind]
            w_unwind = self._get_trade_weights_for_t(holdings_unwind, t_unwind)
            u_unwind_pre = sum(holdings_unwind) * w_unwind * self.leverage_per_trade
            u_unwind_scaled = u_unwind_pre * self._cum_returns_to_scale_unwind(
                t_unwind, t
            )

            u -= u_unwind_scaled

        return u

    def _get_trade_weights_for_t(self, holdings: pd.Series, t: dt.datetime):
        n_short = round(self.forecast_returns.shape[1] * self.percent_short)
        n_long = round(self.forecast_returns.shape[1] * self.percent_long)

        prediction = values_in_time(self.forecast_returns, t)
        prediction_sorted = prediction.sort_values()

        short_trades = prediction_sorted.index[:n_short]
        long_trades = prediction_sorted.index[-n_long:]

        w = pd.Series(0.0, index=prediction.index)
        w[short_trades] = -1.0
        w[long_trades] = 1.0

        w /= sum(abs(w))

        return w

    def _cum_returns_to_scale_unwind(self, t_unwind: dt.datetime, t: dt.datetime):
        df = self.optimizer.actual["return"] + 1
        df = df[(df.index >= t_unwind) & (df.index < t)]

        return df.cumprod().iloc[-1]



from investos.portfolio.constraint_model.base_constraint import BaseConstraint


class MaxWeightConstraint(BaseConstraint):
    """
    A constraint that enforces a limit on the weight of each asset in a portfolio.

    Parameters
    ----------
    limit : float, optional
        The maximum weight of each asset in the portfolio. Defaults to 0.05.

    **kwargs :
        Additional keyword arguments.

    Methods
    -------
    weight_expr(self, t, w_plus, z, v):
        Returns a series of holding constraints based on the portfolio weights after trades.

    """

    def __init__(self, limit: float = 0.025, **kwargs):
        self.limit = limit

    def weight_expr(self, t, w_plus, z, v):
        """
        Returns a series of holding constraints.

        Parameters
        ----------
        t : datetime
            The current time.

        w_plus : series
            Portfolio weights after trades z.

        z : series
            Trades for period t

        v : float
            Value of portfolio at period t

        Returns
        -------
        series
            The holding constraints based on the portfolio weights after trades.
        """
        return w_plus[:-1] <= self.limit


class MinWeightConstraint(BaseConstraint):
    """
    A constraint that enforces a limit on the weight of each asset in a portfolio.

    Parameters
    ----------
    limit : float, optional
        The minimum weight of each asset in the portfolio. Defaults to -0.05.

    **kwargs :
        Additional keyword arguments.

    Methods
    -------
    weight_expr(self, t, w_plus, z, v):
        Returns a series of holding constraints based on the portfolio weights after trades.

    """

    def __init__(self, limit: float = -0.025, **kwargs):
        self.limit = limit

    def weight_expr(self, t, w_plus, z, v):
        """
        Returns a series of holding constraints.

        Parameters
        ----------
        t : datetime
            The current time.

        w_plus : series
            Portfolio weights after trades z.

        z : series
            Trades for period t

        v : float
            Value of portfolio at period t

        Returns
        -------
        series
            The holding constraints based on the portfolio weights after trades.
        """
        return w_plus[:-1] >= self.limit



# http://web.cvxr.com/cvx/doc/basics.html#constraints
class BaseConstraint:
    """
    Base class for constraint objects used in convex portfolio optimization strategies.

    Subclass `BaseConstraint`, and create your own `weight_expr` method to create custom constraints.
    """

    def __init__(self, **kwargs):
        self.optimizer = None  # Set during Optimizer initialization

    def weight_expr(self, t, w_plus, z, v):
        raise NotImplementedError

    def metadata_dict(self):
        metadata_dict = {}

        if getattr(self, "limit", None):
            metadata_dict["limit"] = self.limit

        return metadata_dict



import copy
from functools import wraps

import pandas as pd


def deep_dict_merge(default_d, update_d):
    "Deep copies update_d onto default_d recursively"

    default_d = copy.deepcopy(default_d)
    update_d = copy.deepcopy(update_d)

    def deep_dict_merge_inner(default_d, update_d):
        for k in update_d.keys():
            if (
                k in default_d
                and isinstance(default_d[k], dict)
                and isinstance(update_d[k], dict)
            ):
                deep_dict_merge_inner(default_d[k], update_d[k])
            else:
                default_d[k] = update_d[k]

    deep_dict_merge_inner(default_d, update_d)
    return default_d  # With update_d values copied onto it


def values_in_time(obj, t, tau=None):
    """
    From CVXPortfolio:

    Obtain value(s) of object at time t, or right before.

    Optionally specify time tau>=t for which we want a prediction,
    otherwise it is assumed tau = t.

    obj: callable, pd.Series, pd.DataFrame, or something else.

        If a callable, we return obj(t,tau).

        If obj has an index attribute,
        we try to return obj.loc[t],
        or obj.loc[t, tau], if the index is a MultiIndex.
        If not available, we return obj.

        Otherwise, we return obj.

    t: np.Timestamp (or similar). Time at which we want
        the value.

    tau: np.Timestamp (or similar), or None. Time tau >= t
        of the prediction,  e.g., tau could be tomorrow, t
        today, and we ask for prediction of market volume tomorrow,
        made today. If None, then it is assumed tau = t.

    """

    if callable(obj):
        return obj(t, tau)

    if isinstance(obj, pd.Series) or isinstance(obj, pd.DataFrame):
        try:
            if isinstance(obj.index, pd.MultiIndex):
                return obj.loc[(t, tau)]
            else:
                return obj.loc[t]
        except KeyError:
            return obj

    return obj


def clip_for_dates(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        pd_obj = func(self, *args, **kwargs)
        return pd_obj[
            (pd_obj.index >= self.start_date) & (pd_obj.index <= self.end_date)
        ]

    return wrapper



from investos.portfolio.constraint_model.base_constraint import *
from investos.portfolio.constraint_model.leverage_constraint import *
from investos.portfolio.constraint_model.long_constraint import *
from investos.portfolio.constraint_model.trade_constraint import *
from investos.portfolio.constraint_model.weight_constraint import *



import investos.portfolio

__version__ = "0.2.2"

import os

api_key = os.environ.get("FORECASTOS_API_KEY", "")
api_endpoint = os.environ.get(
    "FORECASTOS_API_ENDPOINT", "https://app.forecastos.com/api/v1"
)



import cvxpy as cvx

from investos.portfolio.constraint_model.base_constraint import BaseConstraint


class MaxLeverageConstraint(BaseConstraint):
    """
    A constraint that enforces a limit on the (absolute) leverage of the portfolio.

    E.g. For leverage of 2.0x, a portfolio with 100MM net value
    (i.e. the portfolio value if it were converted into cash,
    ignoring liquidation / trading costs)
    could have 200MM of (combined long and short) exposure.

    Parameters
    ----------
    limit : float, optional
        The minimum weight of each asset in the portfolio. Defaults to -0.05.

    **kwargs :
        Additional keyword arguments.
    """

    def __init__(self, limit: float = 2.0, **kwargs):
        self.limit = limit

    def weight_expr(self, t, w_plus, z, v):
        """
        Returns a series of holding constraints.

        Parameters
        ----------
        t : datetime
            The current time.

        w_plus : series
            Portfolio weights after trades z.

        z : series
            Trades for period t

        v : float
            Value of portfolio at period t

        Returns
        -------
        series
            The holding constraints based on the portfolio leverage after trades.
        """
        return cvx.sum(cvx.abs(w_plus[:-1])) <= self.limit



import datetime as dt

import cvxpy as cvx
import pandas as pd

import investos.util as util
from investos.portfolio.constraint_model import (
    BaseConstraint,
    EqualLongShortConstraint,
    MaxLeverageConstraint,
    MaxWeightConstraint,
    MinWeightConstraint,
)
from investos.portfolio.cost_model import BaseCost, HoldingCost, TradingCost
from investos.portfolio.risk_model import BaseRisk, StatFactorRisk
from investos.portfolio.strategy import BaseStrategy
from investos.util import values_in_time


class SPO(BaseStrategy):
    """Optimization strategy that builds trade list using single period optimization."""

    BASE_SOLVER_OPTS = {
        "max_iter": 50_000,
        "eps_rel": 0.0000000001,
        "eps_abs": 0.0000000001,
    }

    def __init__(
        self,
        costs: [BaseCost] = [TradingCost(), HoldingCost()],
        constraints: [BaseConstraint] = [
            MinWeightConstraint(),
            MaxWeightConstraint(),
            MaxLeverageConstraint(),
            EqualLongShortConstraint(),
        ],
        risk_model: BaseRisk = StatFactorRisk(),
        solver=cvx.OSQP,
        solver_opts=None,
    ):
        self.forecast_returns = None  # Set by Backtester in init
        self.optimizer = None  # Set by Backtester in init

        self.costs = costs
        self.risk_model = risk_model
        self.constraints = constraints
        self.solver = solver
        self.solver_opts = util.deep_dict_merge(
            self.BASE_SOLVER_OPTS, solver_opts or {}
        )

    def generate_trade_list(self, holdings: pd.Series, t: dt.datetime) -> pd.Series:
        """Calculates and returns trade list (in units of currency passed in) using convex (single period) optimization.

        Parameters
        ----------
        holdings : pandas.Series
            Holdings at beginning of period `t`.
        t : datetime.datetime
            The datetime for associated holdings `holdings`.
        """

        if t is None:
            t = dt.datetime.today()

        value = sum(holdings)
        w = holdings / value  # Portfolio weights
        z = cvx.Variable(w.size)  # Portfolio trades
        wplus = w.values + z  # Portfolio weights after trades

        alpha_term = cvx.sum(
            cvx.multiply(values_in_time(self.forecast_returns, t).values, wplus)
        )

        assert alpha_term.is_concave()

        costs, constraints = [], []

        for cost in self.costs:
            cost_expr, const_expr = cost.weight_expr(t, wplus, z, value)
            costs.append(cost_expr)
            constraints += const_expr

        constraints += [
            item
            for item in (
                con.weight_expr(t, wplus, z, value) for con in self.constraints
            )
        ]

        # For help debugging:

        # for el in costs:
        #     if not el.is_convex():
        #         print(t, el, "is not convex")
        # assert (el.is_convex())

        # for el in constraints:
        # if not el.is_dcp():
        #     print(t, el, "is not dcp")
        # assert (el.is_dcp())

        objective = cvx.Maximize(alpha_term - cvx.sum(costs))
        constraints += [cvx.sum(z) == 0]
        self.prob = cvx.Problem(
            objective, constraints
        )  # Trades need to 0 out, i.e. cash account must adjust to make everything net to 0

        try:
            self.prob.solve(solver=self.solver, **self.solver_opts)

            if self.prob.status == "unbounded":
                print(f"The problem is unbounded at {t}.")
                return self._zerotrade(holdings)

            if self.prob.status == "infeasible":
                print(f"The problem is infeasible at {t}.")
                return self._zerotrade(holdings)

            # print("CVX problem at ", t, self.prob)
            u = pd.Series(index=holdings.index, data=(z.value * value))

            return u

        except (cvx.SolverError, cvx.DCPError, TypeError):
            print(f"The solver failed for {t}.")
            return self._zerotrade(holdings)



from investos.backtest import Result


class ForecastResult(Result):
    """To be implemented.

    For forecast results (where future prices are unknown). Used to generate foward-looking trades / ideal positions.
    """

    def __init__(self):
        pass



import datetime as dt

import cvxpy as cvx
import numpy as np
import pandas as pd

from investos.portfolio.cost_model import BaseCost
from investos.util import values_in_time


class HoldingCost(BaseCost):
    """Calculates cost for holding short positions, given customizable short_rate."""

    def __init__(self):
        self.optimizer = None  # Set during Optimizer initialization

        super().__init__()

    def _estimated_cost_for_optimization(self, t, w_plus, z, value):
        """Estimated holding costs.

        Used by optimization strategy to determine trades.

        Not used to calculate simulated holding costs for backtest performance.
        """
        w_plus = w_plus[:-1]

        self.expression = self._get_short_rate(t) * cvx.neg(
            w_plus
        )  # Borrowing rate for period * negative values

        return cvx.sum(self.expression), []

    def actual_cost(self, t: dt.datetime, h_plus: pd.Series, u: pd.Series) -> pd.Series:
        """Method that calculates per-period (short position) holding costs given period `t` holdings and trades.

        Parameters
        ----------
        t : datetime.datetime
            The datetime for associated trades `u` and holdings plus trades `h_plus`.
        h_plus : pandas.Series
            Holdings at beginning of period t, plus trades for period `t` (`u`). Same as `u` + `h` for `t`.
        u : pandas.Series
            Trades (as values) for period `t`.
        """
        short_rate = self._get_short_rate(t)
        self.last_cost = -np.minimum(0, h_plus.iloc[:-1]) * short_rate

        return sum(self.last_cost)

    def _get_short_rate(self, t):
        # Check if t is period 0
        idx_t = self.optimizer.forecast["return"].index.get_loc(t)
        if idx_t == 0:
            idx_t = 1  # If it is, use first period as proxy

        t_minus_1 = self.optimizer.forecast["return"].index[idx_t - 1]
        fraction_of_year = (t - t_minus_1) / dt.timedelta(365, 0, 0, 0)

        return (
            1 + values_in_time(self.optimizer.config["borrowing"]["short_rate"], t)
        ) ** (fraction_of_year) - 1



import investos.portfolio.constraint_model
import investos.portfolio.cost_model
import investos.portfolio.result
import investos.portfolio.risk_model
import investos.portfolio.strategy
from investos.portfolio.controller import Controller



import datetime as dt

import cvxpy as cvx
import numpy as np
import pandas as pd

from investos.portfolio.cost_model import BaseCost
from investos.util import values_in_time


class TradingCost(BaseCost):
    """Calculates per period cost for trades `u`, based on spread, standard deviation, volume, and price.

    Attributes
    ----------
    sensitivity_coeff : float
        For scaling transaction cost; 1 assumes 1 period's volume moves price by 1 std_dev in vol
    """

    def __init__(self, price_movement_sensitivity=1):
        self.optimizer = None  # Set during Optimizer initialization
        # For scaling realized transaction cost; 1 assumes 1 day's volume moves price by 1 day's vol
        # --> Eventually support DF for this
        self.sensitivity_coeff = price_movement_sensitivity

        super().__init__()

    def _estimated_cost_for_optimization(self, t, w_plus, z, value):
        """Estimated trading costs.

        Used by optimization strategy to determine trades.
        """
        z = z[:-1]
        constraints = []

        # Calculate terms for estimated trading cost
        std_dev = values_in_time(self.optimizer.forecast["std_dev"], t)
        volume_dollars = values_in_time(
            self.optimizer.forecast["volume"], t
        ) * values_in_time(self.optimizer.forecast["price"], t)
        percent_volume_traded_pre_trade_weight = (
            np.abs(value) / volume_dollars
        )  # Multiplied (using cvx) by trade weight (z) below!

        price_movement_term = (
            self.sensitivity_coeff * std_dev * percent_volume_traded_pre_trade_weight
        )
        # END calculate terms for estimated trading cost

        # Create estimated cost expression
        try:  # Spread (estimated) costs
            self.estimate_expression = cvx.multiply(
                values_in_time(self.optimizer.forecast["half_spread"], t), cvx.abs(z)
            )
        except TypeError:
            self.estimate_expression = cvx.multiply(
                values_in_time(self.optimizer.forecast["half_spread"], t).values,
                cvx.abs(z),
            )

        try:  # Price movement due to volume (estimated) costs
            self.estimate_expression += cvx.multiply(
                price_movement_term, cvx.abs(z) ** 2
            )
        except TypeError:
            self.estimate_expression += cvx.multiply(
                price_movement_term.values, cvx.abs(z) ** 2
            )
        # END create estimated cost expression

        return cvx.sum(self.estimate_expression), constraints

    def actual_cost(self, t: dt.datetime, h_plus: pd.Series, u: pd.Series) -> pd.Series:
        """Method that calculates `t` period cost for trades `u`.

        Trading cost (per asset) is assumed to be half of bid-ask spread, plus (per asset) standard deviation * % of period volume traded.
        Assumption is trading 100% of period volume moves price by 1 (default value for `self.sensitivity_coeff`) standard deviation.

        Parameters
        ----------
        t : datetime.datetime
            The datetime for associated trades `u` and holdings plus trades `h_plus`.
        h_plus : pandas.Series
            Holdings at beginning of period t, plus trades for period `t` (`u`). Same as `u` + `h` for `t`.
        u : pandas.Series
            Trades (as values) for period `t`.
        """

        u_no_cash = u.iloc[:-1]

        spread_cost = np.abs(u_no_cash) * values_in_time(
            self.optimizer.actual["half_spread"], t
        )
        std_dev = values_in_time(self.optimizer.actual["std_dev"], t)
        volume_dollars = values_in_time(
            self.optimizer.actual["volume"], t
        ) * values_in_time(self.optimizer.actual["price"], t)
        percent_volume_traded = np.abs(u_no_cash) / volume_dollars

        trading_costs = spread_cost + (
            self.sensitivity_coeff * std_dev * np.abs(u_no_cash) * percent_volume_traded
        )

        self.optimizer.backtest.save_data("costs_trading", t, trading_costs)

        return trading_costs.sum()
